"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    var desc = Object.getOwnPropertyDescriptor(m, k);
    if (!desc || ("get" in desc ? !m.__esModule : desc.writable || desc.configurable)) {
      desc = { enumerable: true, get: function() { return m[k]; } };
    }
    Object.defineProperty(o, k2, desc);
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.allAckedChangesPerspective = exports.RemoteObliteratePerspective = exports.LocalDefaultPerspective = exports.createLocalReconnectingPerspective = exports.LocalReconnectingPerspective = exports.PriorPerspective = void 0;
const constants_js_1 = require("./constants.js");
const mergeTreeNodes_js_1 = require("./mergeTreeNodes.js");
const segmentInfos_js_1 = require("./segmentInfos.js");
const opstampUtils = __importStar(require("./stamps.js"));
class PerspectiveBase {
    isSegmentPresent(seg) {
        if ((0, segmentInfos_js_1.isInserted)(seg) && !this.hasOccurred(seg.insert)) {
            return false;
        }
        if ((0, segmentInfos_js_1.isRemoved)(seg) && seg.removes.some((remove) => this.hasOccurred(remove))) {
            return false;
        }
        return true;
    }
}
/**
 * A perspective which includes edits at or before some reference sequence number alongside all edits from some particular client.
 *
 * @remarks
 * This works for both the local client as well as remote clients since refSeq-based checks disallow unacked edits, but the clientId check
 * catches unacked edits from the local client.
 */
class PriorPerspective extends PerspectiveBase {
    constructor(refSeq, clientId) {
        super();
        this.refSeq = refSeq;
        this.clientId = clientId;
    }
    hasOccurred(stamp) {
        const predatesViaRefSeq = (0, mergeTreeNodes_js_1.seqLTE)(stamp.seq, this.refSeq);
        const predatesViaSameClient = stamp.clientId === this.clientId;
        return predatesViaRefSeq || predatesViaSameClient;
    }
}
exports.PriorPerspective = PriorPerspective;
/**
 * A perspective which includes edits which were either:
 * - acked and at or before some reference sequence number
 * - unacked, but at or before some local sequence number
 *
 * This is a useful perspective when the local client is in the process of reconnecting, since it must
 * rederive positions for unacked ops while only considering a portion of its own edits as having been applied.
 */
class LocalReconnectingPerspective extends PerspectiveBase {
    constructor(refSeq, clientId, localSeq) {
        super();
        this.refSeq = refSeq;
        this.clientId = clientId;
        this.localSeq = localSeq;
    }
    hasOccurred(stamp) {
        const predatesViaRefSeq = (0, mergeTreeNodes_js_1.seqLTE)(stamp.seq, this.refSeq);
        const predatesViaLocalSeq = stamp.localSeq !== undefined && stamp.localSeq <= this.localSeq;
        return predatesViaRefSeq || predatesViaLocalSeq;
    }
}
exports.LocalReconnectingPerspective = LocalReconnectingPerspective;
/**
 * A perspective which includes edits which were either:
 * - acked and at or before some reference sequence number
 * - unacked, but at or before some local sequence number
 *
 * @internal
 */
function createLocalReconnectingPerspective(refSeq, clientId, localSeq) {
    return new LocalReconnectingPerspective(refSeq, clientId, localSeq);
}
exports.createLocalReconnectingPerspective = createLocalReconnectingPerspective;
/**
 * A perspective which includes all known edits.
 *
 * This is the perspective that the application sees.
 * @remarks
 * This can be represented using {@link PriorPerspective} with a refSeq of `Number.MAX_SAFE_INTEGER`, but having an explicit
 * variant of this perspective renders extra refSeq checks unnecessary and is a bit easier to read.
 */
class LocalDefaultPerspective extends PerspectiveBase {
    constructor(clientId) {
        super();
        this.clientId = clientId;
        this.refSeq = Number.MAX_SAFE_INTEGER;
    }
    hasOccurred(_stamp) {
        return true;
    }
}
exports.LocalDefaultPerspective = LocalDefaultPerspective;
/**
 * A perspective dictating whether segments are 'visible' to a remote obliterate operation.
 *
 * NOTE: Beware that partial lengths doesn't support this perspective, in the sense that consulting partial lengths' for the length of a block
 * can give different results than summing the lengths of present segments in that block.
 * This ends up not affecting the current obliterate implementation (which has some special casing in the mapRange calls it uses),
 * but use with caution.
 */
class RemoteObliteratePerspective extends PerspectiveBase {
    constructor(clientId) {
        super();
        this.clientId = clientId;
        this.refSeq = Number.MAX_SAFE_INTEGER;
    }
    hasOccurred(stamp) {
        // Local-only removals are not visible to an obliterate operation, since this means the local removal was concurrent
        // to a remote obliterate and we may need to mark the segment appropriately to reflect this overlapping remove.
        // Every other type of operation is visible: obliterates do not affect segments that have already been removed and acked,
        // and they always affect segments within their range that have not been removed, even if those segments were inserted
        // after the obliterate's refSeq.
        if (isRemoveOperationStamp(stamp) && opstampUtils.isLocal(stamp)) {
            return false;
        }
        return true;
    }
}
exports.RemoteObliteratePerspective = RemoteObliteratePerspective;
function isRemoveOperationStamp(stamp) {
    const { type } = stamp;
    return type === "setRemove" || type === "sliceRemove";
}
exports.allAckedChangesPerspective = new PriorPerspective(Number.MAX_SAFE_INTEGER, constants_js_1.NonCollabClient);
//# sourceMappingURL=perspective.js.map